#if defined(__APPLE_CC__)
#include <OpenGL/gl.h>
#else
#include <GL/gl.h>
#endif

#include "sys.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#define NBPTS_BASE 30	// doit etre pair
#define NBPTS_HIGH 40
#define NBPTS (NBPTS_BASE*NBPTS_HIGH)

GLfloat vertex[NBPTS*3];
unsigned int NBFACES;
GLuint *faces;	// Quads

unsigned endFaceIdx;
void calcCyl() {
	unsigned h, b, p=0;
	NBFACES = NBPTS-NBPTS_BASE;
	endFaceIdx = NBFACES-NBPTS_BASE;
	faces = malloc(4*NBFACES*sizeof(GLuint));
	for (h=0; h<NBPTS_HIGH; h++) {
		for (b=0; b<NBPTS_BASE; b++) {
			double a = (b*M_PI*2./NBPTS_BASE);
			vertex[p*3] = (1.+h*.1)*cos(a);
			vertex[p*3+2] = (-1.-h*.1)*sin(a);
			vertex[p*3+1] = h*1.;
			p++;
		}
	}
	for (p=0; p<NBFACES; p++) {
		int psuiv = 1;
		faces[4*p] = p;
		if (!((p+1)%NBPTS_BASE)) psuiv = 1-NBPTS_BASE;
		faces[4*p+1] = p+psuiv;
		faces[4*p+2] = p+NBPTS_BASE+psuiv;
		faces[4*p+3] = p+NBPTS_BASE;
	}
}

void scrollCyl() {
	unsigned p;
	// 1. les points en bas de l'anneau sont en fait les premiers points
	for (p=0; p<NBPTS_BASE; p++) {
		faces[4*(p+endFaceIdx)+0] = faces[4*(p+endFaceIdx)+3]; //p;
		faces[4*(p+endFaceIdx)+1] = faces[4*(p+endFaceIdx)+2]; //(p<NBPTS_BASE-1?p+1:0);
		faces[4*(p+endFaceIdx)+2] = faces[4*p+2];
		faces[4*(p+endFaceIdx)+3] = faces[4*p+3];
		// relinker le premier anneau
		faces[4*p+2] = faces[4*(p+endFaceIdx)+1];
		faces[4*p+3] = faces[4*(p+endFaceIdx)+0];
	}
	// 2. les points en haut de l'anneau sont les memes, mais changent de position
	for (p=0; p<NBPTS_BASE; p++) {
		vertex[3*faces[4*(p+endFaceIdx)+0]+0] = vertex[3*p+0];
		vertex[3*faces[4*(p+endFaceIdx)+0]+1] = 2;//(vertex[3*p+1]+vertex[3*faces[4*(p+endFaceIdx)+3] +1])*.5;
		vertex[3*faces[4*(p+endFaceIdx)+0]+2] = vertex[3*p+2];
		// a prioris il suffit de changer ceux la...
	}
	endFaceIdx -= NBPTS_BASE;
	if (0==endFaceIdx) endFaceIdx = NBFACES-NBPTS_BASE;
}

void calcBandes() {
	unsigned h, b, p=0;
	NBFACES = (NBPTS_BASE/2)*(NBPTS_HIGH-1);
	faces = malloc(4*NBFACES*sizeof(GLuint));
	for (h=0; h<NBPTS_HIGH; h++) {
		for (b=0; b<NBPTS_BASE; b++) {
			double a = (b*M_PI*2./NBPTS_BASE);
			vertex[p*3] = (1.+h*.1)*cos(a);
			vertex[p*3+2] = (-1.-h*.1)*sin(a);
			vertex[p*3+1] = h*1.;
			p++;
		}
	}
	for (p=0; p<NBFACES; p++) {
		faces[4*p] = p*2;
		faces[4*p+1] = p*2+1;
		faces[4*p+2] = p*2+NBPTS_BASE+1;
		faces[4*p+3] = p*2+NBPTS_BASE;
	}
}

void calcObj() {
	calcCyl();
	//calcBandes();
}

GLfloat camera[16] = {
	1,0,0,0,
	0,1,0,0,
	0,0,1,0,
	0,-17,-25,1	/* decaler en X sur la gauche, vent vers la droite */
};
double angle=0;

#define RESX 20
#define RESY 50
#define RESZ 20

float vent[2][RESX][RESY][RESZ][3];
unsigned ventIdx = 0;

#define nbPerturb 80
struct {
	unsigned life;
	float vx,vy,vz;
	unsigned x,y,z;
} perturb[nbPerturb];

void initVent() {
	unsigned x,y,z;
	for (x=0; x<RESX; x++)
		for (y=0; y<RESY; y++)
			for (z=0; z<RESZ; z++) {
				vent[ventIdx][x][y][z][0]=0;
				vent[ventIdx][x][y][z][1]=0;
				vent[ventIdx][x][y][z][2]=0;
			}
	for (x=0; x<nbPerturb; x++)
		perturb[x].life=0;
}

#define posVx -40.
#define posVy -1.1
#define posVz -40.
#define sizeVx 80.
#define sizeVy 60.
#define sizeVz 80.
#define scaleVx (sizeVx/RESX)
#define scaleVy (sizeVy/RESY)
#define scaleVz (sizeVz/RESZ)

void anim() {
	static int count=50;
	unsigned p;
	unsigned x,y,z;
	//camera[14]-=.2;
	double ca = cos(angle);
	double sa = sin(angle);
	camera[0] = ca;
	camera[2] = sa;
	camera[8] = -sa;
	camera[10] = ca;
	angle += .015;
	/* scroller le mesh */
	if (0==count) {
		scrollCyl();
		count=10;
	}
	count--;
	/* pousser le mesh */
	for (p=0; p<NBPTS; p++) {
		unsigned vx,vy,vz;
		vx = (vertex[p*3+0]-posVx)/scaleVx;
		vy = (vertex[p*3+1]-posVy)/scaleVy;
		vz = (vertex[p*3+2]-posVz)/scaleVz;
		if (vx<RESX && vy<RESY && vz<RESZ) {
			vertex[p*3+0] += vent[ventIdx][vx][vy][vz][0];
			vertex[p*3+1] += vent[ventIdx][vx][vy][vz][1];
			vertex[p*3+2] += vent[ventIdx][vx][vy][vz][2];
		}
	}
	/* ajouter du souffle */
	for (p=0; p<nbPerturb; p++) {
		if (perturb[p].life==0) {
			float yfactor;
			perturb[p].life = 15+(rand()*15.)/RAND_MAX;
			perturb[p].x = 1+((float)rand()*(RESX-2.))/RAND_MAX;
			perturb[p].y = 1+((float)rand()*(RESY-2.))/RAND_MAX;
			perturb[p].z = 1+((float)rand()*(RESZ-2.))/RAND_MAX;
			yfactor = (float)perturb[p].y/RESY;
			perturb[p].vx = -14.*yfactor*((float)perturb[p].z-RESZ/2)/RESZ;
			perturb[p].vy = 1.5*yfactor*((float)rand())/RAND_MAX;
			perturb[p].vz = 14.*yfactor*((float)perturb[p].x-RESX/2)/RESX;
		}
		vent[ventIdx][perturb[p].x][perturb[p].y][perturb[p].z][0] += perturb[p].vx;
		vent[ventIdx][perturb[p].x][perturb[p].y][perturb[p].z][1] += perturb[p].vy;
		vent[ventIdx][perturb[p].x][perturb[p].y][perturb[p].z][2] += perturb[p].vz;
		perturb[p].life--;
	}
	/* dissiper le vent */
	for (x=1; x<RESX-1; x++) {
		for (y=1; y<RESY-1; y++) {
			for (z=1; z<RESZ-1; z++) {
				float aten = 6.;
				vent[ventIdx^1][x][y][z][0] = (
					vent[ventIdx][x-1][y][z][0] + vent[ventIdx][x+1][y][z][0] +
					vent[ventIdx][x][y-1][z][0] + vent[ventIdx][x][y+1][z][0] +
					vent[ventIdx][x][y][z-1][0] + vent[ventIdx][x][y][z+1][0]
						) / aten;
				vent[ventIdx^1][x][y][z][1] = (
					vent[ventIdx][x-1][y][z][1] + vent[ventIdx][x+1][y][z][1] +
					vent[ventIdx][x][y-1][z][1] + vent[ventIdx][x][y+1][z][1] +
					vent[ventIdx][x][y][z-1][1] + vent[ventIdx][x][y][z+1][1]
						) / aten;
				vent[ventIdx^1][x][y][z][2] = (
					vent[ventIdx][x-1][y][z][2] + vent[ventIdx][x+1][y][z][2] +
					vent[ventIdx][x][y-1][z][2] + vent[ventIdx][x][y+1][z][2] +
					vent[ventIdx][x][y][z-1][2] + vent[ventIdx][x][y][z+1][2]
						) / aten;
			}
		}
	}
	ventIdx^=1;
}

GLfloat light[4] = { 1,-1,0,0 };

void display() {
	static unsigned char proj=0;
	static unsigned char wire=0;
	static unsigned char repere=0;
	static unsigned char vbox=0;
	static unsigned char vfield=0;
	static unsigned shadow=1;
	static GLfloat *normales;
	static GLfloat *shad;
	if (!proj) {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glFrustum(-1, 1, -1,1, 1, 1000);
		glMatrixMode(GL_MODELVIEW);
		normales = malloc(3*NBPTS*sizeof(GLfloat));
		shad = malloc(3*NBPTS*sizeof(GLfloat));
		// light
		glLightfv(GL_LIGHT0, GL_POSITION, light);
		glEnable(GL_LIGHT0);
		glEnable(GL_BLEND);
		proj=1;
	}
	glLoadMatrixf(camera);
	SYS_efface(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glDisable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glBegin(GL_QUADS);
	glColor3f(0.1, 0.2, 0.03);
	glVertex3f(-500, -2, -500);
	glVertex3f( 500, -2, -500);
	glVertex3f( 500, -2,  500);
	glVertex3f(-500, -2,  500);
	glEnd();
	glEnable(GL_BLEND);
	if (shadow) {
		unsigned p;
		for (p=0; p<NBPTS; p++) {
			shad[3*p+0] = vertex[3*p+0]-vertex[3*p+1];
			shad[3*p+1] = 0;
			shad[3*p+2] = vertex[3*p+2];
		}
		glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, shad);
		glColor4f(.03,.1,.01,.9);
		glDrawElements(GL_QUADS, NBFACES*4, GL_UNSIGNED_INT, (void*)faces);
	}
	glBlendFunc(GL_ONE, GL_ONE);
	glEnable(GL_LIGHTING);
	if (wire) {
		unsigned f;
		glColor3f(1,1,1);
		for (f=0; f<NBFACES; f++) {
			glBegin(GL_LINES);
			glVertex3f(vertex[faces[f*4]*3], vertex[faces[f*4]*3+1],vertex[faces[f*4]*3+2]);
			glVertex3f(vertex[faces[f*4+1]*3], vertex[faces[f*4+1]*3+1],vertex[faces[f*4+1]*3+2]);
			glVertex3f(vertex[faces[f*4+1]*3], vertex[faces[f*4+1]*3+1],vertex[faces[f*4+1]*3+2]);
			glVertex3f(vertex[faces[f*4+2]*3], vertex[faces[f*4+2]*3+1],vertex[faces[f*4+2]*3+2]);
			glVertex3f(vertex[faces[f*4+2]*3], vertex[faces[f*4+2]*3+1],vertex[faces[f*4+2]*3+2]);
			glVertex3f(vertex[faces[f*4+3]*3], vertex[faces[f*4+3]*3+1],vertex[faces[f*4+3]*3+2]);
			glVertex3f(vertex[faces[f*4+3]*3], vertex[faces[f*4+3]*3+1],vertex[faces[f*4+3]*3+2]);			
			glVertex3f(vertex[faces[f*4]*3], vertex[faces[f*4]*3+1],vertex[faces[f*4]*3+2]);
			glEnd();
		}
	} else {
		unsigned f;
		for (f=0; f<NBFACES; f++) {
			unsigned v = faces[4*f+0];
			GLfloat ax = vertex[3*faces[4*f+1]+0] - vertex[3*faces[4*f+0]+0];
			GLfloat ay = vertex[3*faces[4*f+1]+1] - vertex[3*faces[4*f+0]+1];
			GLfloat az = vertex[3*faces[4*f+1]+2] - vertex[3*faces[4*f+0]+2];
			GLfloat bx = vertex[3*faces[4*f+3]+0] - vertex[3*faces[4*f+0]+0];
			GLfloat by = vertex[3*faces[4*f+3]+1] - vertex[3*faces[4*f+0]+1];
			GLfloat bz = vertex[3*faces[4*f+3]+2] - vertex[3*faces[4*f+0]+2];
			GLfloat nx = ay*bz-az*by;
			GLfloat ny = az*bx-ax*bz;
			GLfloat nz = ax*by-ay*bx;
			GLfloat n = nx*nx+ny*ny+nz*nz;
			assert(v<NBPTS);
			normales[3*v+0] = nx/(n+1);
			normales[3*v+1] = ny/(n+1);
			normales[3*v+2] = nz/(n+1);
		}
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, vertex);
		glNormalPointer(GL_FLOAT, 0, normales);
		glColor4f(1,1,1,.9001);
		glDrawElements(GL_QUADS, NBFACES*4, GL_UNSIGNED_INT, (void*)faces);
	}
	if (repere) {
		glBegin(GL_LINES);
		glColor3f(1,1,1);
		glVertex4f(0,0,0,1);
		glVertex4f(1,0,0,1);
		glVertex4f(0,0,0,1);
		glVertex4f(0,1,0,1);
		glVertex4f(0,0,0,1);
		glVertex4f(0,0,1,1);
		glEnd();
	}
	if (vbox) {
		glBegin(GL_LINES);
		glColor3f(.5,.5,1);
		glVertex3f(posVx, posVy, posVz);
		glVertex3f(posVx+RESX*scaleVx, posVy, posVz);
		glVertex3f(posVx, posVy, posVz);
		glVertex3f(posVx, posVy+RESY*scaleVy, posVz);
		glVertex3f(posVx, posVy, posVz);
		glVertex3f(posVx, posVy, posVz+RESZ*scaleVz);
		glVertex3f(posVx+RESX*scaleVx, posVy+RESY*scaleVy, posVz+RESZ*scaleVz);
		glVertex3f(posVx+RESX*scaleVx, posVy+RESY*scaleVy, posVz);
		glVertex3f(posVx+RESX*scaleVx, posVy+RESY*scaleVy, posVz+RESZ*scaleVz);
		glVertex3f(posVx, posVy+RESY*scaleVy, posVz+RESZ*scaleVz);
		glVertex3f(posVx+RESX*scaleVx, posVy+RESY*scaleVy, posVz+RESZ*scaleVz);
		glVertex3f(posVx+RESX*scaleVx, posVy, posVz+RESZ*scaleVz);
		glEnd();
	}
	if (vfield) {
		unsigned x, y, z;
		glBegin(GL_LINES);
		glColor3f(.5,.5,.5);
		for (x=1; x<RESX-1; x+=2)
			for (y=1; y<RESY-1; y+=2)
				for (z=1; z<RESZ-1; z+=2) {
#define VFSCALE 5.
					glVertex3f(posVx+x*scaleVx, posVy+y*scaleVy, posVz+z*scaleVz);
					glVertex3f(posVx+x*scaleVx+vent[ventIdx][x][y][z][0]*VFSCALE, posVy+y*scaleVy+vent[ventIdx][x][y][z][1]*VFSCALE, posVz+z*scaleVz+vent[ventIdx][x][y][z][2]*VFSCALE);
				}
		glColor3f(1,.3,.3);
		for (x=0; x<nbPerturb; x++) {
			glVertex3f(posVx+perturb[x].x*scaleVx, posVy+perturb[x].y*scaleVy, posVz+perturb[x].z*scaleVz);
			glVertex3f(posVx+perturb[x].x*scaleVx+perturb[x].vx*VFSCALE, posVy+perturb[x].y*scaleVy+perturb[x].vy*VFSCALE, posVz+perturb[x].z*scaleVz+perturb[x].vz*VFSCALE);
		}
		glEnd();
	}
	anim();
	SYS_affiche();
}

#define TIMER 25

void monTimer(int no) {
	SYS_reaffiche();
	SYS_timer(TIMER, &monTimer, 1);
}


int main(int nbarg, char **args) {
	calcObj();
	initVent();
	SYS_displayFunc = &display;
	SYS_mouseFunc = NULL;
	SYS_motionFunc = NULL;
	SYS_keyFunc = NULL;
	SYS_keyUpFunc = NULL;
	SYS_idleFunc = NULL;
	SYS_init(&nbarg, args);
	SYS_timer(TIMER, &monTimer, 1);
	SYS_run();
	return 666;
}
